//////////////////////////////// Fragment shader
//Modified version of dota2 shader from substance designer
#version 120

#include "../common/common.glsl"

#define M_INV_PI 0.31830988618379067153776752674503

// #define ALG_NORMAL_DIRECTX
#define ALG_NORMAL_OPENGL

#ifdef ALG_NORMAL_DIRECTX
  // #define FLIP_NORMAL_X
  #define FLIP_NORMAL_Y
  // #define FLIP_NORMAL_Z
#endif //#ifdef ALG_NORMAL_DIRECTX

#ifdef ALG_NORMAL_OPENGL
  // #define FLIP_NORMAL_X
  // #define FLIP_NORMAL_Y
  // #define FLIP_NORMAL_Z
#endif //#ifdef ALG_NORMAL_OPENGL

varying vec3 iFS_Normal;
varying vec2 iFS_UV;
varying vec3 iFS_Tangent;
varying vec3 iFS_Binormal;
varying vec3 iFS_PointWS;



uniform bool bVertexLit = false;
uniform sampler2D Diffuse;
uniform bool bUseDiffuse = true;
uniform sampler2D Normal;
uniform bool bUseNormalMap = true;
uniform bool bFlipGreen = false;
uniform sampler2D EnvmapMask;
uniform sampler2D SpecularMask;
uniform sampler2D GlossinessMap;
uniform bool bUseGlossiness = false;
uniform sampler2D MetallicMap;
uniform bool bUseMetallic = false;
uniform float fMetallicBoost = 1.0f;
uniform float fMaskContrast = 1.0f;
uniform vec3 fEnvmapColor = vec3(1.0f,1.0f,1.0f);
uniform bool bUseEnvmap = true;
uniform float fEnvmapBlur = 0;
uniform float fEnvmapFresnel = 0;
uniform vec3 fSpecularColor = vec3(1.0f,1.0f,1.0f);
uniform float fSpecularExponent = 5.0f;
uniform float fSpecularScale = 1.0f;
uniform vec3 fSpecularRanges = vec3(0.25, 0.5, 1.0);
uniform vec3 fFresnelRanges = vec3(0.5, 0.5, 1.0);
uniform bool bCustomFresnelRangesSpec = false;
uniform vec3 fAmbientColor = vec3(0.0f,0.0f,0.0f);
uniform float fAmbientScale = 0.5f;

uniform vec3 Lamp0Pos = vec3(0.0,0.0,70.0);
uniform vec3 Lamp0Color = vec3(1.0,1.0,1.0);
uniform float Lamp0Intensity = 1.0;
uniform vec3 Lamp1Pos = vec3(70.0,0.0,0.0);
uniform vec3 Lamp1Color = vec3(0.198,0.198,0.198);
uniform float Lamp1Intensity = 1.0;

uniform sampler2D environmentMap;
uniform float envRotation = 0.0f;

uniform vec3 shCoefs[10];

uniform mat4 matWorldViewProjection;
uniform mat4 matWorldInverseTranspose;
uniform mat4 matViewInverse;
uniform mat4 matWorld;

#ifdef FLIP_NORMAL_X
  result.x = -result.x;
#endif // ifdef FLIP_NORMAL_X
#ifdef FLIP_NORMAL_Y
  result.y = -result.y;
#endif // ifdef FLIP_NORMAL_Y
#ifdef FLIP_NORMAL_Z
  result.z = -result.z;
#endif // ifdef FLIP_NORMAL_Z

vec3 normalVecOSToWS(vec3 normal)
{
  return normal;
}

vec3 linearToSrgb(vec3 c)
{
    return pow(c,vec3(0.45,0.45,0.45));
}

float saturate(float v)
{
  return clamp(v, 0.0, 1.0);
}

vec3 saturate(vec3 v)
{
  return clamp(v, 0.0, 1.0);
}

struct lightProp
{
  vec3 lightVector;
  vec3 lightColor;
};

vec3 samplePanorama(sampler2D map, vec3 dir, float lod)
{
	float n = length(dir.xz);
	vec2 pos = vec2( (n>0.0000001) ? dir.x / n : 0.0, dir.y);
	pos = acos(pos)*M_INV_PI;
	pos.x = (dir.z > 0.0) ? pos.x*0.5 : 1.0-(pos.x*0.5);
	pos.y = 1.0-pos.y;
        return texture2DLod(map, pos, lod).rgb;
}

vec3 irradianceFromSH(vec3 n)
{
	return (shCoefs[0]*n.x + shCoefs[1]*n.y + shCoefs[2]*n.z + shCoefs[3])*n.x
		+ (shCoefs[4]*n.y + shCoefs[5]*n.z + shCoefs[6])*n.y
		+ (shCoefs[7]*n.z + shCoefs[8])*n.z
		+ shCoefs[9];
}

// If ( x * Tangent, y * Binormal, z * Normal ) is used, normal maps need swizzle
vec3 swizzle(vec3 ref)
{
  return vec3(ref.y, ref.x, ref.z); // swizzle it
}

// Cubemaps need different lookup, 3dsmax is Z up
vec3 cubeSwizzle(vec3 ref)
{
  return vec3(ref.x,ref.z,ref.y);
}

float Fresnel(vec3 N, vec3 V)
{
  float fresnel = 1.0 - saturate( dot( N, V ) );
  return fresnel * fresnel;
}


float Fresnel(vec3 N, vec3 V, vec3 fresnelRanges)
{
  float result, f = Fresnel(N, V);

  if (f > 0.5f)
    result = mix(fresnelRanges.g, fresnelRanges.b, (2*f) - 1.0);
  else
    result = mix(fresnelRanges.r, fresnelRanges.g, 2*f);

  return result;
}

void ambientTerm(vec3 Normal, out vec3 finalDiffuse)
{
  vec3 ambient = fAmbientColor * fAmbientScale;

  // ambient = textureCube(Environment, vec4(Normal,9)) * fAmbientScale;
  ambient = irradianceFromSH(rotate(Normal,envRotation)) * fAmbientScale;

  vec3 diffuseLight = ambient;

  finalDiffuse = diffuseLight; // Output our diffuse lighting
}

void diffuseTerm(vec3 Normal, vec3 LightVector, vec3 LightColor, out vec3 finalDiffuse)
{
  // Normalize
  LightVector = normalize(LightVector);

  float NdotL = saturate(dot(Normal, LightVector));
  float halfLambert = NdotL * 0.5 + 0.5;

  vec3 diffuseLight = vec3(0.0f, 0.0f, 0.0f);

  diffuseLight += NdotL;
  diffuseLight *= LightColor; // Modulate with light color
  finalDiffuse = diffuseLight; // Output our diffuse lighting
}

void specularTerm(vec3 Normal, vec3 LightVector, vec3 LightColor, vec3 ViewVector, float SpecularExponent, out vec3 finalSpecular)
{

  float NdotL = dot(Normal, LightVector);
  float NdotH = dot(Normal, normalize(LightVector + ViewVector));

  NdotL = saturate(NdotL);
  NdotH = saturate(NdotH);
  float specularPower = pow(NdotH, SpecularExponent);

  specularPower *= NdotL;
  vec3 specularLight = LightColor;
  specularLight *= specularPower;
  specularLight *= fSpecularScale;

  finalSpecular = vec3(0.0f, 0.0f, 0.0f); // to avoid uninitialzed variable
  finalSpecular += specularLight; // Output our final specular lighting
}

void envmapTerm(vec3 Normal, vec3 ViewVector, out vec3 finalEnvmap )
{
  finalEnvmap = vec3(0.0f, 0.0f, 0.0f);

  vec3 Ln = -reflect( ViewVector, Normal);
  finalEnvmap += samplePanorama(environmentMap, rotate(Ln,envRotation), 1 + fEnvmapBlur);
}

void main()
{

  vec3 cameraPosWS = matViewInverse[3].xyz;
  vec3 pointToLight0DirWS = Lamp0Pos - iFS_PointWS;
	float pointToLight0Length = length(pointToLight0DirWS);
	pointToLight0DirWS *= 1.0 / pointToLight0Length;
	vec3 pointToLight1DirWS = Lamp1Pos - iFS_PointWS;
	float pointToLight1Length = length(Lamp1Pos - iFS_PointWS);
	pointToLight1DirWS *= 1.0 / pointToLight1Length;
	vec3 pointToCameraDirWS = normalize(cameraPosWS - iFS_PointWS);
  vec3 Lamp0 = Lamp0Color * Lamp0Intensity * lampAttenuation(pointToLight0Length);
  vec3 Lamp1 = Lamp1Color * Lamp1Intensity * lampAttenuation(pointToLight1Length);

  vec2 texCoord = iFS_UV;
  vec3 V = normalize(cameraPosWS - iFS_PointWS);

  vec3 fixedNormalWS = iFS_Normal;  // HACK for empty normal textures
	vec3 normalTS = texture2D(Normal,texCoord).xyz;
	if(length(normalTS)>0.0001)
	{
		normalTS = fixNormalSample(normalTS, true);
		fixedNormalWS = normalize( normalTS.x*iFS_Tangent + normalTS.y*iFS_Binormal + normalTS.z*iFS_Normal );
	}

  vec3 N = fixedNormalWS;

  vec3 mask1 = texture2D(EnvmapMask, texCoord).rgb;
  vec3 mask2 = texture2D(SpecularMask, texCoord).rgb;

  if (!bVertexLit)
    mask1 = mask2;

  float fresnel = Fresnel(N, V);
  float fresnelRanges = Fresnel(N, V, fSpecularRanges);

  float flSpecularExponent = texture2D(GlossinessMap, texCoord).r;

  vec3 albedo = texture2D(Diffuse, texCoord).rgb;

  vec3 diffuseColor = vec3(0.5f, 0.5f, 0.5f); // global diffuse color

  if (bUseDiffuse)
  {
    diffuseColor = srgb_to_linear(albedo); // assign the diffuse texture to our global diffuse color
  }

  vec3 diffuseLighting;
  vec3 ambientLighting;
  vec3 specularLighting;
  vec3 envmapLighting;
  vec3 finalEnvmap = vec3(0.0f, 0.0f, 0.0f);
  vec3 finalDiffuse = vec3(0.0f, 0.0f, 0.0f);
  vec3 finalSpecular = vec3(0.0f, 0.0f, 0.0f);

  float specExponent = fSpecularExponent;
  float flSpecMask = pow(albedo.r, fMaskContrast);

  if (bUseGlossiness && bVertexLit) {
		specExponent = 1.0 - flSpecularExponent + 150.0 * flSpecularExponent;
	}

  ambientTerm( N, ambientLighting);
  finalDiffuse += ambientLighting;

  diffuseTerm( N, pointToLight0DirWS, Lamp0, diffuseLighting );
  finalDiffuse += diffuseLighting;
  diffuseTerm( N, pointToLight1DirWS, Lamp1, diffuseLighting );
  finalDiffuse += diffuseLighting;

  vec3 specNormal = N;
  if (!bVertexLit && !bUseEnvmap)
    specNormal = iFS_Normal;

  specularTerm( specNormal, pointToLight0DirWS, Lamp0, V, specExponent, specularLighting);
  finalSpecular += specularLighting;
  specularTerm( specNormal, pointToLight1DirWS, Lamp1, V, specExponent, specularLighting);
  finalSpecular += specularLighting;

  finalSpecular *= mask2.r;
  finalSpecular *= srgb_to_linear(fSpecularColor);

  if(!bVertexLit) {
    finalSpecular *= flSpecMask;
  }

  if(bUseEnvmap)
  {
    envmapTerm( N, V, envmapLighting );
    finalEnvmap += linearToSrgb(envmapLighting);
  }

  finalEnvmap *= srgb_to_linear(fEnvmapColor);
  finalEnvmap *= mask1;

  vec3 cMetalSpecular = finalSpecular * albedo * fMetallicBoost;
  vec3 cMetalEnvmap = finalEnvmap;

  if (bUseMetallic && bVertexLit) {
    finalSpecular = mix(finalSpecular, cMetalSpecular, texture2D(MetallicMap, texCoord).r);
    cMetalEnvmap = mix(finalEnvmap, finalEnvmap * albedo * fMetallicBoost, texture2D(MetallicMap, texCoord).r);
  }

  if (bVertexLit) {
    finalSpecular *= fresnelRanges;
    finalEnvmap = mix(finalEnvmap, cMetalEnvmap * fresnelRanges, fEnvmapFresnel);
  } else {
    finalEnvmap = mix(finalEnvmap, finalEnvmap * fresnel, fEnvmapFresnel);
  }

  finalDiffuse *= diffuseColor;

  vec3 final = (finalDiffuse) + (finalSpecular) + (finalEnvmap);
  //vec3 final = vec3(1.0f, 1.0f, 1.0f) * fresnelRanges;

  gl_FragColor = vec4(final, 1.0f);
}
